@@ -21,6 +21,7 @@ open Printf
 open Benl_core
 open Benl_base
 open Tyxml.Html
+open Benl_compat.Html
 
 let use_colors = ref false
 let output_file = ref None
@@ -36,6 +37,19 @@ let output_format_from_string = function
   | "color" -> use_colors := true; Text
   | _ -> Levels
 
+let format_of_string = function
+  | "html" -> Xhtml
+  | "json" -> Json
+  | "txt" -> Text
+  | "levels" -> Levels
+  | s -> Printf.ksprintf invalid_arg "invalid format: %s" s
+
+let string_of_format = function
+  | Text -> "txt"
+  | Json -> "json"
+  | Xhtml -> "html"
+  | Levels -> "levels"
+
 let map_fun f l =
   let fr = Benl_frontend.get_selected_frontend () in
   if fr.Benl_frontend.name = "monitor" then
@@ -77,10 +91,10 @@ let format_arch x =
 
 let ben_webpage = "https://ben.debian.net"
 
-let print_dep_line src deps =
-  printf "%s:" !!!src;
-  S.iter (fun dep -> printf " %s" !!!dep) deps;
-  printf "\n%!"
+let print_dep_line fmt src deps =
+  Format.fprintf fmt "%s:" !!!src;
+  S.iter (fun dep -> Format.fprintf fmt " %s" !!!dep) deps;
+  Format.fprintf fmt "\n%!"
 
 let spec = Arg.align [
   "--run-debcheck", Arg.Set Benl_data.run_debcheck, " Run debcheck to register installability information";
@@ -159,7 +173,7 @@ let archs_list config =
     with _ -> []
   in
   let archs_list = Benl_core.uniq (release_archs_list @ ignored_archs_list) in
-  List.sort Pervasives.compare archs_list
+  List.sort Benl_compat.compare archs_list
 
 let compute_monitor_data config sources binaries rounds =
   map_fun begin fun xs ->
@@ -192,8 +206,22 @@ let compute_monitor_data config sources binaries rounds =
     end packages
   end rounds
 
-let print_text_monitor config sources binaries rounds =
-  let monitor_data = compute_monitor_data config sources binaries rounds in
+type transition_data = {
+    config : Benl_types.expr Benl_core.StringMap.t;
+    monitor_data :
+      ([ `source ] Package.t *
+         (string * (string * Benl_base.status)) list)
+        list list;
+    sources : ([ `source ], [ `source ] Package.t) Package.Map.t;
+    binaries : [ `binary ] Package.t Benl_modules.PAMap.t;
+    rounds : [ `source ] Package.Name.t list list;
+    dep_graph : ([ `source ], [ `source ] Package.Set.t) Package.Map.t;
+    all : int;
+    bad : int;
+    packages : [ `source ] Benl_data.S.t;
+  }
+
+let print_text_monitor fmt {config; monitor_data; sources; rounds; _} =
   let nmax = M.fold begin fun src _ accu ->
     let n = String.length !!!src in
     if n > accu then n else accu
@@ -218,7 +246,7 @@ let print_text_monitor config sources binaries rounds =
     Scanf.format_from_string (Buffer.contents buf) "%d"
   in
   List.iteri begin fun i xs ->
-    printf header_fmt i;
+    Format.fprintf fmt header_fmt i;
     List.iter begin fun (src, states) ->
       let in_testing =
         try not (Package.get "is-in-testing" src = "no")
@@ -226,17 +254,16 @@ let print_text_monitor config sources binaries rounds =
       in
       let sname = Package.get "package" src in
       let sname = if in_testing then sname else "_"^sname in
-      printf "%*s:" (nmax+3) sname;
+      Format.fprintf fmt "%*s:" (nmax+3) sname;
       List.iter begin fun (arch, (_, state)) ->
-        printf " %s" (format_arch state arch)
+        Format.fprintf fmt " %s" (format_arch state arch)
       end states;
-      printf "\n";
+      Format.fprintf fmt "\n";
     end xs;
-    printf "\n"
+    Format.fprintf fmt "\n"
   end monitor_data
 
-let print_json_monitor config sources binaries rounds =
-  let monitor_data = compute_monitor_data config sources binaries rounds in
+let print_json_monitor fmt {monitor_data; _} =
   let levels = List.mapi begin fun i xs ->
     let srcs = List.map begin fun (src, states) ->
       let in_testing =
@@ -252,7 +279,7 @@ let print_json_monitor config sources binaries rounds =
     end xs in
     Printf.sprintf "  \"%d\" : {\n%s\n  }" i (String.concat ",\n" srcs)
   end monitor_data in
-  Printf.printf "{\n%s\n}" (String.concat ",\n" levels)
+  Format.fprintf fmt "{\n%s\n}" (String.concat ",\n" levels)
 
 let a_link url text =
   a ~a:[a_href (uri_of_string url)] [pcdata text]
@@ -280,7 +307,8 @@ let changelog t ver dir src =
 
 let generated_on_text () =
   [ pcdata "Page generated by ";
-    a_link ben_webpage "Ben";
+    a_link ben_webpage "Ben ";
+    pcdata Benl_version.version;
     pcdata (Printf.sprintf " on %s" (Benl_core.get_rfc2822_date ()))
   ]
 
@@ -343,13 +371,13 @@ let cut_head text head =
   with _ -> text
 
 let beautify_text =
-  let r_link = Pcre.regexp "#[0-9]{4,}|[a-z]{3,}://[^\\s><]+|<[^\\s><]+@[^\\s><]+>|[Pp][Tt][Ss]:[a-z0-9+\\-\\.]+|[Bb][Uu][Ii][Ll][Dd][Dd]:[a-z0-9+\\-\\.]+" in
+  let r_link = Re.Pcre.regexp "#[0-9]{4,}|[a-z]{3,}://[^\\s><]+|<[^\\s><]+@[^\\s><]+>|[Pp][Tt][Ss]:[a-z0-9+\\-\\.]+|[Bb][Uu][Ii][Ll][Dd][Dd]:[a-z0-9+\\-\\.]+" in
   fun tpl text ->
-    let t = Pcre.full_split ~rex:r_link text in
+    let t = Re.Pcre.full_split ~rex:r_link text in
     List.map
       (function
-        | Pcre.Text s -> pcdata s
-        | Pcre.Delim s ->
+        | Re.Pcre.Text s -> pcdata s
+        | Re.Pcre.Delim s ->
           let l = String.lowercase_ascii s in
           if s.[0] = '#' then
             let ss = String.sub s 1 (String.length s -1) in
@@ -369,7 +397,7 @@ let beautify_text =
             a_link link s
           else
             a_link s s
-        | Pcre.Group _ | Pcre.NoGroup -> (* Ignore this case *) pcdata ""
+        | Re.Pcre.Group _ | Re.Pcre.NoGroup -> (* Ignore this case *) pcdata ""
       )
       t
 
@@ -391,7 +419,7 @@ let compute_transition_data data config =
   let rounds, sources, binaries, dep_graph = compute_graph data config in
   let monitor_data = compute_monitor_data config sources binaries rounds in
   let all, bad, packages = generate_stats monitor_data in
-  monitor_data, sources, binaries, dep_graph, all, bad, packages
+  {config; monitor_data; sources; binaries; rounds; dep_graph; all; bad; packages}
 
 let has_testing_data monitor_data =
   match monitor_data with
@@ -402,7 +430,8 @@ let has_testing_data monitor_data =
      with Not_found -> false)
   | _ -> false
 
-let print_html_monitor config template monitor_data binaries dep_graph packages has_testing_data extra =
+let print_html_monitor template {config; monitor_data; binaries; dep_graph; packages; _} extra =
+  let has_testing_data = has_testing_data monitor_data in
   let affected = packages in
   let mytitle =
     try
@@ -455,18 +484,23 @@ let print_html_monitor config template monitor_data binaries dep_graph packages
       [
         pcdata "Filter by status: ";
         input ~a:[a_input_type `Checkbox; a_id "good"] ();
-        pcdata "good ";
+        label ~a:[a_label_for "good"] [pcdata "good"];
+        pcdata " ";
         input ~a:[a_input_type `Checkbox; a_checked (); a_id "partial"] ();
-        pcdata "partial ";
+        label ~a:[a_label_for "partial"] [pcdata "partial"];
+        pcdata " ";
         input ~a:[a_input_type `Checkbox; a_checked (); a_id "bad"] ();
-        pcdata "bad ";
-        input ~a:[a_input_type `Checkbox; a_checked (); a_id "unknown"] (); pcdata "unknown";
+        label ~a:[a_label_for "bad"] [pcdata "bad"];
+        pcdata " ";
+        input ~a:[a_input_type `Checkbox; a_checked (); a_id "unknown"] ();
+        label ~a:[a_label_for "unknown"] [pcdata "unknown"];
+        pcdata " ";
         span ~a:[a_id "count"] [];
       ];
     div (if has_testing_data
       then [
         input ~a:[a_input_type `Checkbox; a_id "notintesting"] ();
-        pcdata "ignore packages that are not in testing";
+        label ~a:[a_label_for "notintesting"] [pcdata "ignore packages that are not in testing"];
       ]
       else []
     );
@@ -611,16 +645,26 @@ let print_html_monitor config template monitor_data binaries dep_graph packages
   in
   html
 
-let print_dependency_levels dep_graph rounds =
+let print_dependency_levels fmt {dep_graph; rounds; _} =
   List.iteri begin fun i xs ->
-    printf "===[ Dependency level %d ]=====\n" i;
+    Format.fprintf fmt "===[ Dependency level %d ]=====\n" i;
     let packages = List.sort (fun x y -> compare !!!x !!!y) xs in
     List.iter begin fun src ->
       let deps = M.find src dep_graph in
-      print_dep_line src deps
+      print_dep_line fmt src deps
     end packages
   end rounds
 
+let print_monitor format fmt transition_data =
+  match format with
+  | Levels -> print_dependency_levels fmt transition_data
+  | Text -> print_text_monitor fmt transition_data
+  | Json -> print_json_monitor fmt transition_data
+  | Xhtml ->
+     let template = Benl_templates.get_registered_template () in
+     let output = print_html_monitor template transition_data None in
+     Format.fprintf fmt "%a\n%!" (Tyxml.Html.pp ()) output
+
 let main _ =
   let config = match !input_source with
     | Benl_types.NoSource -> Benl_error.raise Benl_error.Missing_configuration_file
@@ -631,35 +675,22 @@ let main _ =
     (Benl_clflags.get_config config "architectures")
   in
   let data = Benl_data.load_cache archs_list in
-  let rounds, sources, binaries, dep_graph =
-    compute_graph data config in
-  match !output_format with
-    | Levels -> print_dependency_levels dep_graph rounds
-    | Text -> print_text_monitor config sources binaries rounds
-    | Json -> print_json_monitor config sources binaries rounds
-    | Xhtml ->
-      let template = Benl_templates.get_registered_template () in
-      let monitor_data = compute_monitor_data config sources binaries rounds in
-      let _, _, packages = generate_stats monitor_data in
-      let has_testing_data = has_testing_data monitor_data in
-      let output =
-        print_html_monitor
-          config
-          template
-          monitor_data
-          binaries
-          dep_graph
-          packages
-          has_testing_data
-          None
-      in
-      match !output_file with
-      | None ->
-          Format.printf "%a" (Tyxml.Html.pp ()) output;
-          print_newline ()
-      | Some file ->
-          check_media_dir (Filename.basename file);
-          Benl_utils.dump_xhtml_to_file file output
+  let transition_data = compute_transition_data data config in
+  let fmt, close =
+    match !output_file with
+    | None -> Format.std_formatter, (fun () -> ())
+    | Some file ->
+       let oc = open_out file in
+       Format.formatter_of_out_channel oc,
+       (fun () -> close_out oc)
+  in
+  let () =
+    match !output_format, !output_file with
+    | Xhtml, Some file -> check_media_dir (Filename.basename file)
+    | _, _ -> ()
+  in
+  print_monitor !output_format fmt transition_data;
+  close ()
 
 let anon_fun file =
   if Benl_core.ends_with file ".ben" then
