@@ -18,6 +18,7 @@
 (**************************************************************************)
 
 open Tyxml.Html
+open Benl_compat.Html
 open Printf
 open Benl_core
 open Benl_clflags
@@ -32,6 +33,7 @@ let global_config = ref (FilePath.concat !config_dir "global.conf")
 let lock = ref "ben.lock"
 let clean = ref true
 let tconfig = ref None
+let output_formats = ref [Ben_monitor.Xhtml]
 
 open Benl_types
 open Benl_modules
@@ -79,6 +81,10 @@ let read_global_config () =
           Ben_monitor.baseurl := url
         | "template", (EString template) ->
           Benl_templates.load_template template;
+        | "output-formats", formats ->
+           output_formats :=
+             to_string_l "output-formats" formats
+             |> List.map Ben_monitor.format_of_string
         | item, _ ->
             warn (Unknown_configuration_item item)
     )
@@ -177,16 +183,14 @@ let read_transition_config file =
   (* Read a .ben file *)
   transition, (Benl_frontend.read_ben_file file)
 
+let get_export config =
+  try Benl_clflags.get_config config "export" = Benl_types.Etrue
+  with _ -> true
+
 let get_transition_data data name config =
   try
     let transition_data = Ben_monitor.compute_transition_data data config in
-    let monitor_data, _, _, _, _, _, _ = transition_data in
-    let has_testing_data = Ben_monitor.has_testing_data monitor_data in
-    let export =
-      try
-        Benl_clflags.get_config config "export" = Benl_types.Etrue
-      with _ -> true in
-    Some (export, transition_data, has_testing_data)
+    Some transition_data
   with e ->
     Benl_error.warn_exn ("Failed to process transition " ^ name) e;
     None
@@ -205,8 +209,8 @@ let smerge _ v1 v2 = match v1, v2 with
   | None   , Some v2 -> Some v2
   | _                -> None
 
-let html_path_t name =
-  Filename.concat "html" (Printf.sprintf "%s.html" name)
+let format_path_t format name =
+  Filename.concat format (Printf.sprintf "%s.%s" name format)
 
 let print_html_collisions (hits : (SMap.key * SMap.key list) list) =
   let hits = List.fold_left
@@ -242,9 +246,7 @@ let print_html_collisions (hits : (SMap.key * SMap.key list) list) =
     | h::l -> ul (h::l)
   ]
 
-let print_html_monitor config template file transition_data has_testing_data collisions =
-  let monitor_data, _, binaries, dep_graph, _, _, packages = transition_data in
-  let ($) = Filename.concat in
+let print_html_monitor fmt template file transition_data collisions =
   let (!!) = Filename.basename in
   let collisions_div =
     try
@@ -255,29 +257,43 @@ let print_html_monitor config template file transition_data has_testing_data col
   in
   let output =
     Ben_monitor.print_html_monitor
-      config
       template
-      monitor_data
-      binaries
-      dep_graph
-      packages
-      has_testing_data
+      transition_data
       collisions_div
   in
-  let htmlp = html_path_t !!file in
+  Format.fprintf fmt "%a\n%!" (Tyxml.Html.pp ()) output
+
+let print_monitor_inner fmt format template file transition_data collisions =
+  match format with
+  | Ben_monitor.Xhtml -> print_html_monitor fmt template file transition_data collisions
+  | _ -> Ben_monitor.print_monitor format fmt transition_data
+
+let print_monitor format template file transition_data collisions =
+  let ($) = Filename.concat in
+  let (!!) = Filename.basename in
+  let format_name = Ben_monitor.string_of_format format in
+  let htmlp = format_path_t format_name !!file in
   let html = !base $ htmlp in
   p "Generating %s\n" htmlp;
   try
-    Benl_utils.dump_xhtml_to_file html output
+    let oc = open_out html in
+    begin
+      try
+        let fmt = Format.formatter_of_out_channel oc in
+        print_monitor_inner fmt format template file transition_data collisions;
+        close_out oc
+      with e ->
+        close_out oc; Benl_compat.raise e
+    end
   with e ->
-    Benl_error.warn_exn ("Failed to generate" ^ html) e
+    Benl_error.warn_exn ("Failed to generate " ^ html) e
 
 let compute_collisions results =
   let data_map = List.fold_left
     (fun
       data_map
-      (_, transition, _, (_, transition_data, _)) ->
-        let  _, _, _, _, _, _, pkgs = transition_data in
+      (transition, _, transition_data) ->
+        let pkgs = transition_data.Ben_monitor.packages in
         let new_data = Benl_data.S.fold
           (fun package data ->
             SMap.add (Package.Name.to_string package) [transition] data
@@ -308,9 +324,11 @@ let compute_collisions results =
 let generate_stats results =
   List.fold_left
     (fun (packages, profiles)
-      (p, t, export, transition_data, _) ->
-        let _, _, _, _, all, bad, pkgs = transition_data in
-        let htmlp = html_path_t t in
+      (t, p, transition_data) ->
+        let Ben_monitor.{config; all; bad; _} = transition_data in
+        let export = get_export config in
+        let pkgs = transition_data.Ben_monitor.packages in
+        let htmlp = format_path_t "html" t in
         let profiles = sadd
           profiles
           (string_of_profile p)
@@ -419,7 +437,7 @@ let tracker template profiles =
         with _ ->
           List.assoc Unknown profiles_desc
       in
-      let tlist = List.sort Pervasives.compare tlist in
+      let tlist = List.sort Benl_compat.compare tlist in
       let tlist = List.map (tget show_score) tlist in
       match tlist with
         | [] -> acc
@@ -443,18 +461,18 @@ let tracker template profiles =
   with exn ->
     Benl_error.error_exn "Failed to generate index.html" exn
 
-let () = at_exit (fun () ->
-  try
-    rm [lockf ()]
-  with _ -> ()
-)
-
 let main () =
   let () = read_global_config () in
   let lockf = lockf () in
   if test Exists lockf then
     eprintf "Please wait until %s is removed!\n" lockf
   else
+    let () = at_exit (fun () ->
+      try
+        rm [lockf]
+      with _ -> ()
+      )
+    in
     try
       let lockf_b = Filename.dirname lockf in
       let () = if test Exists lockf_b then
@@ -462,9 +480,13 @@ let main () =
         else
           eprintf "%s doesn't exist. Skipping creation of lock file.\n" lockf_b
       in
-      let htmld = Filename.concat !base "html" in
-      if test (Not Exists) htmld then
-        mkdir ~parent:true htmld;
+      let () =
+        List.iter
+          (fun format ->
+            let dir = Filename.concat !base (Ben_monitor.string_of_format format) in
+            if test (Not Exists) dir then mkdir ~parent:true dir
+          ) !output_formats
+      in
       Ben_monitor.check_media_dir !base;
       let confd = !config_dir in
       let test_cond = And (Size_not_null,
@@ -537,14 +559,14 @@ let main () =
                 (string_of_profile profile)
                 name
             in
-            config, name, profile, (get_transition_data cache name config)
+            name, profile, (get_transition_data cache name config)
           )
           transitions
       in
       let results =
         List.fold_left
           (fun results -> function
-          | (c, n, p, Some d) -> (c, n, p, d) :: results
+          | (n, p, Some d) -> (n, p, d) :: results
           | _ -> results
           )
           []
@@ -554,25 +576,20 @@ let main () =
       let collisions = compute_collisions results in
       (* Generate an HTML page for each transition *)
       let () = Benl_parallel.iter
-        (fun (config, transition, _, (_, transition_data, has_testing_data)) ->
-          print_html_monitor
-            config
-            template
-            transition
-            transition_data
-            has_testing_data
-            collisions
+        (fun (transition, _, transition_data) ->
+          List.iter
+            (fun format ->
+              print_monitor
+                format
+                template
+                transition
+                transition_data
+                collisions
+            ) !output_formats
         )
         results
       in
       (* Generate the packages.yaml file *)
-      let results =
-        Benl_parallel.map
-          (fun (_, t, p, (export, transition_data, has_testing_data)) ->
-            p, t, export, transition_data, has_testing_data
-          )
-          results
-      in
       let packages, profiles = generate_stats results in
       let () = dump_yaml packages "packages.yaml" in
       (* Clean up the HTML directory from old files *)
